from .base import Widget
from .csv import CSV
from .datatables import DataTablesJS
from .reacttable import ReactTable
from .highcharts import HighCharts
from .matplotlib import Matplotlib
from .pandas import Pandas



# noinspection PyUnresolvedReferences
from .database import *
# noinspection PyUnresolvedReferences
from .slicer import *
# noinspection PyUnresolvedReferences
from .slicer.widgets import *

__version__ = '1.4.8'



from .pandas import Pandas


class CSV(Pandas):
    def transform(self, data_frame, slicer, dimensions, references):
        result_df = super(CSV, self).transform(data_frame, slicer, dimensions, references)
        result_df.columns.names = [None]
        return result_df.to_csv()



class SlicerException(Exception):
    pass


class QueryException(SlicerException):
    pass


class MissingTableJoinException(SlicerException):
    pass


class CircularJoinsException(SlicerException):
    pass


class RollupException(SlicerException):
    pass


class MetricRequiredException(SlicerException):
    pass


class ContinuousDimensionRequiredException(SlicerException):
    pass



from pypika import (
    PostgreSQLQuery,
    functions as fn,
    terms,
)
from .base import Database


class DateTrunc(terms.Function):
    """
    Wrapper for the PostgreSQL date_trunc function
    """

    def __init__(self, field, date_format, alias=None):
        super(DateTrunc, self).__init__('DATE_TRUNC', date_format, field, alias=alias)
        # Setting the fields here means we can access the TRUNC args by name.
        self.field = field
        self.date_format = date_format
        self.alias = alias


class PostgreSQLDatabase(Database):
    """
    PostgreSQL client that uses the psycopg module.
    """

    # The pypika query class to use for constructing queries
    query_cls = PostgreSQLQuery

    def __init__(self, host='localhost', port=5432, database=None,
                 user=None, password=None, max_processes=1, cache_middleware=None):
        super(PostgreSQLDatabase, self).__init__(host, port, database,
                                                 max_processes=max_processes,
                                                 cache_middleware=cache_middleware)
        self.user = user
        self.password = password

    def connect(self):
        import psycopg2

        return psycopg2.connect(host=self.host, port=self.port, dbname=self.database,
                                user=self.user, password=self.password)

    def trunc_date(self, field, interval):
        return DateTrunc(field, str(interval))

    def date_add(self, field, date_part, interval):
        return fn.DateAdd(str(date_part), interval, field)



from .base import SlicerElement
from .filters import ComparatorFilter


class Metric(SlicerElement):
    """
    The `Metric` class represents a metric in the `Slicer` object.

    :param alias:
        A unique identifier used to identify the metric when writing slicer queries. This value must be unique over the metrics in the slicer.

    :param definition:
        A pypika expression which is used to select the value when building SQL queries. For metrics, this query **must** be aggregated, since queries always use a ``GROUP BY`` clause an metrics are not used as a group.

    :param label: (optional)
        A display value used for the metric. This is used for rendering the labels within the visualizations. If not set, the alias will be used as the default.

    :param precision: (optional)
        A precision value for rounding decimals. By default, no rounding will be applied.

    :param prefix: (optional)
        A prefix for rendering labels in visualizations such as '$'

    :param suffix:
        A suffix for rendering labels in visualizations such as '€'
    """

    def __init__(self, key, definition, label=None, precision=None, prefix=None, suffix=None):
        super(Metric, self).__init__(key, label, definition)
        self.precision = precision
        self.prefix = prefix
        self.suffix = suffix

    def __eq__(self, other):
        return ComparatorFilter(self.definition, ComparatorFilter.Operator.eq, other)

    def __ne__(self, other):
        return ComparatorFilter(self.definition, ComparatorFilter.Operator.ne, other)

    def __gt__(self, other):
        return ComparatorFilter(self.definition, ComparatorFilter.Operator.gt, other)

    def __ge__(self, other):
        return ComparatorFilter(self.definition, ComparatorFilter.Operator.gte, other)

    def __lt__(self, other):
        return ComparatorFilter(self.definition, ComparatorFilter.Operator.lt, other)

    def __le__(self, other):
        return ComparatorFilter(self.definition, ComparatorFilter.Operator.lte, other)

    def __repr__(self):
        return "slicer.metrics.{}".format(self.key)



from typing import Union

from fireant import (
    Metric,
    Operation,
)
from fireant.slicer.exceptions import MetricRequiredException
from fireant.utils import immutable


class Widget:
    def __init__(self, *items: Union[Metric, Operation]):
        self.items = list(items)

    @immutable
    def item(self, item):
        self.items.append(item)

    @property
    def metrics(self):
        if 0 == len(self.items):
            raise MetricRequiredException(str(self))

        return [metric
                for group in self.items
                for metric in getattr(group, 'metrics', [group])]

    @property
    def operations(self):
        return [item
                for item in self.items
                if isinstance(item, Operation)]

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.items == other.items

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ','.join(str(m) for m in self.items))


class TransformableWidget(Widget):
    # This attribute can be overridden in order to paginate in groups. Useful in cases like for charts where pagination
    # should be applied to the number of series rather than the number of data points.
    group_pagination = False

    def transform(self, data_frame, slicer, dimensions, references):
        """
        - Main entry point -

        Transformers the result set `pd.DataFrame` from a slicer query into the output format for this specific widget
        type.

        :param data_frame:
            The data frame containing the data. Index must match the dimensions parameter.
        :param slicer:
            The slicer that is in use.
        :param dimensions:
            A list of dimensions that are being rendered.
        :param references:
            A list of references that are being rendered.
        :return:
            A dict meant to be dumped as JSON.
        """
        raise NotImplementedError()



import itertools

from .dimensions import DisplayDimension
from .queries import (
    DimensionChoicesQueryBuilder,
    DimensionLatestQueryBuilder,
    SlicerQueryBuilder,
)


class _Container(object):
    """
    This is a list of slicer elements, metrics or dimensions, used for accessing an element by key with a dot syntax.

    Example:

    .. code-block:: python

        slicer = Slicer(
            dimensions=[
                Dimension(key='my_dimension1')
            ]
        )
        slicer.dimensions.my_dimension1
    """
    def __init__(self, items):
        self._items = items
        for item in items:
            setattr(self, item.key, item)

            # Special case to include display definitions for filters
            if item.has_display_field:
                setattr(self, item.display.key, DisplayDimension(item))

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, item):
        return getattr(self, item)

    def __contains__(self, item):
        return hasattr(self, item)

    def __eq__(self, other):
        """
        Checks if the other object is an instance of _Container and has the same number of items with matching keys.
        """
        return isinstance(other, _Container) and \
               all([a is not None
                        and b is not None
                        and a.key == b.key
                        for a, b in itertools.zip_longest(self._items, getattr(other, '_items', ()))])


class Slicer(object):
    """
    WRITEME
    """
    class Dimensions(_Container):
        pass

    class Metrics(_Container):
        pass

    class Fields(_Container):
        pass

    def __init__(self, table, database, joins=(), dimensions=(), metrics=(),
                 hint_table=None,
                 always_query_all_metrics=False):
        """
        Constructor for a slicer.  Contains all the fields to initialize the slicer.

        :param table: (Required)
            A pypika Table reference. The primary table that this slicer will retrieve data from.

        :param database:  (Required)
            A Database reference. Holds the connection details used by this slicer to execute queries.

        :param metrics: (Required: At least one)
            A list of metrics which can be queried.  Metrics are the types of data that are displayed.

        :param dimensions: (Optional)
            A list of dimensions used for grouping metrics.  Dimensions are used as the axes in charts, the indices in
            tables, and also for splitting charts into multiple lines.

        :param joins:  (Optional)
            A list of join descriptions for joining additional tables.  Joined tables are only used when querying a
            metric or dimension which requires it.

        :param hint_table: (Optional)
            A hint table used for querying dimension options.  If not present, the table will be used.  The hint_table
            must have the same definition as the table omitting dimensions which do not have a set of options (such as
            datetime or boolean dimensions) and the metrics.  This is provided to more efficiently query dimension
            options.

        :param always_query_all_metrics: (Default: False)
            When true, all metrics will be included in database queries in order to increase cache hits.
        """
        self.table = table
        self.database = database
        self.joins = joins

        self.hint_table = hint_table
        self.dimensions = Slicer.Dimensions(dimensions)
        self.metrics = Slicer.Metrics(metrics)
        self.fields = Slicer.Fields(metrics + dimensions)

        # add query builder entry points
        self.data = SlicerQueryBuilder(self)
        self.latest = DimensionLatestQueryBuilder(self)
        for dimension in dimensions:
            dimension.choices = DimensionChoicesQueryBuilder(self, dimension)

        self.always_query_all_metrics = always_query_all_metrics

    def __eq__(self, other):
        return isinstance(other, Slicer) \
               and self.metrics == other.metrics \
               and self.dimensions == other.dimensions

    def __repr__(self):
        return 'Slicer(metrics=[{}],dimensions=[{}])' \
            .format(','.join([m.key for m in self.metrics]),
                    ','.join([d.key for d in self.dimensions]))



from pypika import Not
from pypika.functions import Lower


class Filter(object):
    def __init__(self, definition):
        self.definition = definition

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and str(self.definition) == str(other.definition)

    def __repr__(self):
        return str(self.definition)


class DimensionFilter(Filter):
    pass


class MetricFilter(Filter):
    pass


class ComparatorFilter(MetricFilter):
    class Operator(object):
        eq = 'eq'
        ne = 'ne'
        gt = 'gt'
        lt = 'lt'
        gte = 'gte'
        lte = 'lte'

    def __init__(self, metric_definition, operator, value):
        definition = getattr(metric_definition, operator)(value)
        super(ComparatorFilter, self).__init__(definition)


class BooleanFilter(DimensionFilter):
    def __init__(self, element_key, value):
        definition = element_key if value else Not(element_key)
        super(BooleanFilter, self).__init__(definition)


class ContainsFilter(DimensionFilter):
    def __init__(self, dimension_definition, values):
        definition = dimension_definition.isin(values)
        super(ContainsFilter, self).__init__(definition)


class ExcludesFilter(DimensionFilter):
    def __init__(self, dimension_definition, values):
        definition = dimension_definition.notin(values)
        super(ExcludesFilter, self).__init__(definition)


class RangeFilter(DimensionFilter):
    def __init__(self, dimension_definition, start, stop):
        definition = dimension_definition[start:stop]
        super(RangeFilter, self).__init__(definition)


class PatternFilter(DimensionFilter):
    def _apply(self, dimension_definition, patterns):
        definition = Lower(dimension_definition).like(Lower(patterns[0]))

        for pattern in patterns[1:]:
            definition |= Lower(dimension_definition).like(Lower(pattern))

        return definition

    def __init__(self, dimension_definition, pattern, *patterns):
        definition = self._apply(dimension_definition, (pattern,) + patterns)
        super(PatternFilter, self).__init__(definition)


class AntiPatternFilter(PatternFilter):
    def _apply(self, dimension_definition, pattern):
        return super(AntiPatternFilter, self)._apply(dimension_definition, pattern).negate()



import itertools

import pandas as pd
from datetime import timedelta

from fireant import (
    DatetimeDimension,
    formats,
    utils,
)
from .base import TransformableWidget
from .chart_base import (
    ChartWidget,
    ContinuousAxisSeries,
)
from .helpers import (
    dimensional_metric_label,
    extract_display_values,
)
from ..references import (
    reference_key,
    reference_label,
)

DEFAULT_COLORS = (
    "#DDDF0D",
    "#55BF3B",
    "#DF5353",
    "#7798BF",
    "#AAEEEE",
    "#FF0066",
    "#EEAAEE",
    "#DF5353",
    "#7798BF",
    "#AAEEEE",
)

DASH_STYLES = (
    'Solid',
    'Dash',
    'Dot',
    'DashDot',
    'LongDash',
    'LongDashDot',
    'ShortDash',
    'ShortDashDot',
    'LongDashDotDot',
    'ShortDashDotDot',
    'ShortDot',
)

MARKER_SYMBOLS = (
    "circle",
    "square",
    "diamond",
    "triangle",
    "triangle-down",
)

SERIES_NEEDING_MARKER = (ChartWidget.LineSeries, ChartWidget.AreaSeries)
TS_UPPER_BOUND = pd.Timestamp.max - timedelta(seconds=1)


class HighCharts(ChartWidget, TransformableWidget):
    # Pagination should be applied to groups of the 0th index level (the x-axis) in order to paginate series
    group_pagination = True

    def __init__(self, title=None, colors=None, x_axis_visible=True, tooltip_visible=True):
        super(HighCharts, self).__init__()
        self.title = title
        self.colors = colors or DEFAULT_COLORS
        self.x_axis_visible = x_axis_visible
        self.tooltip_visible = tooltip_visible

    def __repr__(self):
        return ".".join(["HighCharts()"] + [repr(axis) for axis in self.items])

    def transform(self, data_frame, slicer, dimensions, references):
        """
        - Main entry point -

        Transforms a data frame into HighCharts JSON format.

        See https://api.highcharts.com/highcharts/

        :param data_frame:
            The data frame containing the data. Index must match the dimensions parameter.
        :param slicer:
            The slicer that is in use.
        :param dimensions:
            A list of dimensions that are being rendered.
        :param references:
            A list of references that are being rendered.
        :return:
            A dict meant to be dumped as JSON.
        """
        colors = itertools.cycle(self.colors)

        dimension_display_values = extract_display_values(dimensions, data_frame)
        render_series_label = dimensional_metric_label(dimensions, dimension_display_values)
        is_timeseries = dimensions and isinstance(dimensions[0], DatetimeDimension)

        # Timestamp.max is used as a marker for rolled up dimensions (totals). Filter out the totals value for the
        # dimension used for the x-axis
        if is_timeseries:
            data_frame = data_frame.loc[:TS_UPPER_BOUND]

        # Group the results by index levels after the 0th, one for each series
        # This will result in a series for every combination of dimension values and each series will contain a data set
        # across the 0th dimension (used for the x-axis)
        series_data_frames = self._group_by_series(data_frame)

        total_num_series = sum([len(axis)
                                for axis in self.items])

        y_axes, series = [], []
        for axis_idx, axis in enumerate(self.items):
            # Tee the colors iterator so we can peek at the next color. The next color in the iterator becomes the
            # axis color. The first series on each axis should match the axis color, and will progress the colors
            # iterator. The next axis color should be the next color in the iterator after the last color used by a
            # series on the current axis in order to get a better variety of color.
            colors, tee_colors = itertools.tee(colors)
            axis_color = next(tee_colors)

            # prepend axes, append series, this keeps everything ordered left-to-right
            y_axes[0:0] = self._render_y_axis(axis_idx,
                                              axis_color if 1 < total_num_series else None,
                                              references)

            series += self._render_series(axis,
                                          axis_idx,
                                          axis_color,
                                          colors,
                                          series_data_frames,
                                          render_series_label,
                                          references,
                                          is_timeseries)

        x_axis = self._render_x_axis(data_frame, dimensions, dimension_display_values)

        return {
            "title": {"text": self.title},
            "xAxis": x_axis,
            "yAxis": y_axes,
            'colors': self.colors,
            "series": series,
            "tooltip": {"shared": True, "useHTML": True, "enabled": self.tooltip_visible},
            "legend": {"useHTML": True},
        }

    def _render_x_axis(self, data_frame, dimensions, dimension_display_values):
        """
        Renders the xAxis configuration.

        https://api.highcharts.com/highcharts/xAxis

        :param data_frame:
        :param dimension_display_values:
        :return:
        """
        first_level = data_frame.index.levels[0] \
            if isinstance(data_frame.index, pd.MultiIndex) \
            else data_frame.index

        if dimensions and isinstance(dimensions[0], DatetimeDimension):
            return {
                "type": "datetime",
                "visible": self.x_axis_visible,
            }

        categories = ["All"] \
            if not isinstance(data_frame.index, pd.MultiIndex) and data_frame.index.name is None \
            else [utils.getdeepattr(dimension_display_values,
                                    (first_level.name, dimension_value),
                                    dimension_value or 'Totals')
                  for dimension_value in first_level]

        return {
            "type": "category",
            "categories": categories,
            "visible": self.x_axis_visible,
        }

    def _group_by_series(self, data_frame):
        if not isinstance(data_frame.index, pd.MultiIndex):
            return [([], data_frame)]

        series = data_frame.index.names[1:]
        return data_frame.groupby(level=series, sort=False)

    def _render_y_axis(self, axis_idx, color, references):
        """
        Renders the yAxis configuration.

        https://api.highcharts.com/highcharts/yAxis

        :param axis_idx:
        :param color:
        :param references:
        :return:
        """
        axis = self.items[axis_idx]

        y_axes = [{
            "id": str(axis_idx),
            "title": {"text": None},
            "labels": {"style": {"color": color}},
            "visible": axis.y_axis_visible,
        }]

        y_axes += [{
            "id": "{}_{}".format(axis_idx, reference.key),
            "title": {"text": reference.label},
            "opposite": True,
            "labels": {"style": {"color": color}},
            "visible": axis.y_axis_visible,
        }
            for reference in references
            if reference.delta]

        return y_axes

    def _render_series(self, axis, axis_idx, axis_color, colors, series_data_frames, render_series_label,
                       references, is_timeseries=False):
        """
        Renders the series configuration.

        https://api.highcharts.com/highcharts/series

        :param axis:
        :param axis_idx:
        :param axis_color:
        :param colors:
        :param series_data_frames:
        :param render_series_label:
        :param references:
        :param is_timeseries:
        :return:
        """
        hc_series = []
        for series in axis:
            symbols = itertools.cycle(MARKER_SYMBOLS)

            for (dimension_values, group_df), symbol in zip(series_data_frames, symbols):
                if is_timeseries:
                    group_df = group_df.sort_index(level=0)

                dimension_values = utils.wrap_list(dimension_values)

                if isinstance(series, self.PieSeries):
                    # pie charts suck
                    for reference in [None] + references:
                        hc_series.append(self._render_pie_series(series,
                                                                 reference,
                                                                 group_df,
                                                                 render_series_label))
                    continue

                # With a single axis, use different colors for each series
                # With multiple axes, use the same color for the entire axis and only change the dash style
                series_color = next(colors)

                for reference, dash_style in zip([None] + references, itertools.cycle(DASH_STYLES)):
                    metric_key = utils.format_metric_key(reference_key(series.metric, reference))

                    hc_series.append({
                        "type": series.type,

                        "name": render_series_label(dimension_values, series.metric, reference),

                        "data": self._render_data(group_df, metric_key, is_timeseries),

                        "tooltip": self._render_tooltip(series.metric),

                        "yAxis": ("{}_{}".format(axis_idx, reference.key)
                                  if reference is not None and reference.delta
                                  else str(axis_idx)),

                        "marker": ({"symbol": symbol, "fillColor": axis_color or series_color}
                                   if isinstance(series, SERIES_NEEDING_MARKER)
                                   else {}),

                        "stacking": series.stacking,
                    })

                    if isinstance(series, ContinuousAxisSeries):
                        # Set each series in a continuous series to a specific color
                        hc_series[-1]["color"] = series_color
                        hc_series[-1]["dashStyle"] = dash_style

        return hc_series

    def _render_pie_series(self, series, reference, data_frame, render_series_label):
        metric = series.metric
        name = reference_label(metric, reference)
        df_key = utils.format_metric_key(series.metric.key)

        data = []
        for dimension_values, y in data_frame[df_key].sort_values(ascending=False).iteritems():
            data.append({
                "name": render_series_label(dimension_values) if dimension_values else name,
                "y": formats.metric_value(y),
            })

        return {
            "name": name,
            "type": series.type,
            "data": data,
            'tooltip': {
                'pointFormat': '<span style="color:{point.color}">\u25CF</span> {series.name}: '
                               '<b>{point.y} ({point.percentage:.1f}%)</b><br/>',
                'valueDecimals': metric.precision,
                'valuePrefix': metric.prefix,
                'valueSuffix': metric.suffix,
            },
        }

    def _render_data(self, group_df, metric_key, is_timeseries):
        if not is_timeseries:
            return [formats.metric_value(y) for y in group_df[metric_key].values]

        series = []
        for dimension_values, y in group_df[metric_key].iteritems():
            first_dimension_value = utils.wrap_list(dimension_values)[0]

            if pd.isnull(first_dimension_value):
                # Ignore totals on the x-axis.
                continue

            series.append((formats.date_as_millis(first_dimension_value),
                           formats.metric_value(y)))

        return series

    def _render_tooltip(self, metric):
        return {
            "valuePrefix": metric.prefix,
            "valueSuffix": metric.suffix,
            "valueDecimals": metric.precision,
        }



import sys
from collections import OrderedDict

import pandas as pd

MAX_TIMESTAMP = pd.Timestamp.max
MAX_NUMBER = sys.maxsize
MAX_STRING = '~~totals'


def wrap_list(value):
    return value if isinstance(value, (tuple, list)) else [value]


def setdeepattr(d, keys, value):
    """
    Similar to the built-in `setattr`, this function accepts a list/tuple of keys to set a value deep in a `dict`

    Given the following dict structure

    .. code-block:: python

        d = {
          'A': {
            '0': {
              'a': 1,
              'b': 2,
            }
          },
        }

    Calling `setdeepattr` with a key path to a value deep in the structure will set that value. If the value or any
    of the objects in the key path do not exist, then a dict will be created.

    .. code-block:: python

        # Overwrites the value in `d` at A.0.a, which was 1, to 3
        setdeepattr(d, ('A', '0', 'a'), 3)

        # Adds an entry in `d` to A.0 with the key 'c' and the value 3
        setdeepattr(d, ('A', '0', 'c'), 3)

        # Adds an entry in `d` with the key 'X' and the value a new dict
        # Adds an entry in `d` to `X` with the key '0' and the value a new dict
        # Adds an entry in `d` to `X.0` with the key 'a' and the value 0
        setdeepattr(d, ('X', '0', 'a'), 0)

    :param d:
        A dict value with nested dict attributes.
    :param keys:
        A list/tuple path of keys in `d` to the desired value
    :param value:
        The value to set at the given path `keys`.
    """
    if not isinstance(keys, (list, tuple)):
        keys = (keys,)

    top, *rest = keys

    if rest:
        if top not in d:
            d[top] = {}

        setdeepattr(d[top], rest, value)

    else:
        d[top] = value


def getdeepattr(d, keys, default_value=None):
    """
    Similar to the built-in `getattr`, this function accepts a list/tuple of keys to get a value deep in a `dict`

    Given the following dict structure

    .. code-block:: python

        d = {
          'A': {
            '0': {
              'a': 1,
              'b': 2,
            }
          },
        }

    Calling `getdeepattr` with a key path to a value deep in the structure will return that value. If the value or any
    of the objects in the key path do not exist, then the default value is returned.

    .. code-block:: python

        assert 1 == getdeepattr(d, ('A', '0', 'a'))
        assert 2 == getdeepattr(d, ('A', '0', 'b'))
        assert 0 == getdeepattr(d, ('A', '0', 'c'), default_value=0)
        assert 0 == getdeepattr(d, ('X', '0', 'a'), default_value=0)

    :param d:
        A dict value with nested dict attributes.
    :param keys:
        A list/tuple path of keys in `d` to the desired value
    :param default_value:
        A default value that will be returned if the path `keys` does not yield a value.
    :return:
        The value following the path `keys` or `default_value`
    """
    d_level = d

    for key in keys:
        if key not in d_level:
            return default_value

        d_level = d_level[key]

    return d_level


def flatten(items):
    return [item for level in items for item in wrap_list(level)]


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    return [l[i:i + n]
            for i in range(0, len(l), n)]


def slice_first(item):
    if isinstance(item, (tuple, list)):
        return item[0]
    return item


def filter_duplicates(iterable):
    filtered_list, seen = [], set()
    for item in iterable:
        key = slice_first(item)

        if key in seen:
            continue

        seen.add(key)
        filtered_list.append(item)

    return filtered_list


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.

    https://stackoverflow.com/questions/38987/how-to-merge-two-python-dictionaries-in-a-single-expression
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def immutable(func):
    """
    Decorator for wrapper "builder" functions.  These are functions on the Query class or other classes used for
    building queries which mutate the query and return self.  To make the build functions immutable, this decorator is
    used which will deepcopy the current instance.  This decorator will return the return value of the inner function
    or the new copy of the instance.  The inner function does not need to return self.
    """
    import copy

    def _copy(self, *args, **kwargs):
        self_copy = copy.deepcopy(self)
        result = func(self_copy, *args, **kwargs)

        # Return self if the inner function returns None.  This way the inner function can return something
        # different (for example when creating joins, a different builder is returned).
        if result is None:
            return self_copy

        return result

    return _copy


def ordered_distinct_list(l):
    seen = set()
    return [x
            for x in l
            if not x in seen
            and not seen.add(x)]


def ordered_distinct_list_by_attr(l, attr='key'):
    seen = set()
    return [x
            for x in l
            if not getattr(x, attr) in seen
            and not seen.add(getattr(x, attr))]


def groupby(items, by):
    """
    Group items using a function to derive a key.

    :param items:
        The items to group

    :param by:
        A lambda function to create a key based on the item

    :return:
        an Ordered dict
    """
    result = OrderedDict()
    for item in items:
        key = by(item)

        if key in result:
            result[key].append(item)
        else:
            result[key] = [item]

    return result


def groupby_first_level(index):
    seen = set()
    return [x[1:]
            for x in list(index)
            if x[1:] not in seen and not seen.add(x[1:])]


def format_key(key, prefix=None):
    if key is None:
        return key

    if prefix is not None:
        return '${}${}'.format(prefix, key)

    return '${}'.format(key)


def format_dimension_key(key):
    return format_key(key, 'd')


def format_metric_key(key):
    return format_key(key, 'm')


def repr_field_key(key):
    field_type_symbol = key[1]
    field_key = key[3:]
    field_type = {'m': 'metrics', 'd': 'dimensions'}[field_type_symbol]
    return 'slicer.{}.{}'.format(field_type, field_key)



from pypika import (
    VerticaQuery,
    functions as fn,
    terms,
)

from .base import Database


class Trunc(terms.Function):
    """
    Wrapper for Vertica TRUNC function for truncating dates.
    """

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('TRUNC', field, date_format, alias=alias)
        # Setting the fields here means we can access the TRUNC args by name.
        self.field = field
        self.date_format = date_format
        self.alias = alias


class VerticaDatabase(Database):
    """
    Vertica client that uses the vertica_python driver.
    """

    # The pypika query class to use for constructing queries
    query_cls = VerticaQuery

    DATETIME_INTERVALS = {
        'hour': 'HH',
        'day': 'DD',
        'week': 'IW',
        'month': 'MM',
        'quarter': 'Q',
        'year': 'Y'
    }

    def __init__(self, host='localhost', port=5433, database='vertica', user='vertica', password=None,
                 read_timeout=None, max_processes=1, cache_middleware=None):
        super(VerticaDatabase, self).__init__(host, port, database,
                                              max_processes=max_processes,
                                              cache_middleware=cache_middleware)
        self.user = user
        self.password = password
        self.read_timeout = read_timeout

    def connect(self):
        import vertica_python

        return vertica_python.connect(host=self.host, port=self.port, database=self.database,
                                      user=self.user, password=self.password,
                                      read_timeout=self.read_timeout,
                                      unicode_error='replace')

    def trunc_date(self, field, interval):
        trunc_date_interval = self.DATETIME_INTERVALS.get(str(interval), 'DD')
        return Trunc(field, trunc_date_interval)

    def date_add(self, field, date_part, interval):
        return fn.TimestampAdd(str(date_part), interval, field)



class NumericInterval(object):
    def __init__(self, size=1, offset=0):
        self.size = size
        self.offset = offset

    def __eq__(self, other):
        return all([isinstance(other, NumericInterval),
                    self.size == other.size,
                    self.offset == other.offset])

    def __hash__(self):
        return hash('NumericInterval %d %d' % (self.size, self.offset))

    def __repr__(self):
        return 'NumericInterval(size=%d,offset=%d)' % (self.size, self.offset)


class DatetimeInterval(object):
    def __init__(self, key):
        self.key = key

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return isinstance(other, DatetimeInterval) \
               and self.key == other.key

    def __str__(self):
        return self.key

    def __repr__(self):
        return 'DatetimeInterval(\'{}\')'.format(self.key)


hourly = DatetimeInterval('hour')
daily = DatetimeInterval('day')
weekly = DatetimeInterval('week')
monthly = DatetimeInterval('month')
quarterly = DatetimeInterval('quarter')
annually = DatetimeInterval('year')



import functools

import pandas as pd
from dateutil.relativedelta import relativedelta

from fireant.slicer.dimensions import DatetimeDimension
from fireant.slicer.filters import RangeFilter
from fireant.slicer.operations import RollingOperation


def adjust_daterange_filter_for_rolling_window(dimensions, operations, filters):
    """
    This function adjusts date filters for a rolling operation in order to select enough date to compute the values for
    within the original range.

    It only applies when using a date dimension in the first position and a RangeFilter is used on that dimension. It
    is meant to be applied to a slicer query.

    :param dimensions:
        The dimensions applied to a slicer query
    :param operations:
        The dimensions used in widgets in a slicer query
    :param filters:
        The filters applied to a slicer query
    :return:
    """
    has_datetime_dimension_in_first_dimension_pos = not len(dimensions) \
                                                    or not isinstance(dimensions[0], DatetimeDimension)
    if has_datetime_dimension_in_first_dimension_pos:
        return filters

    has_rolling = any([isinstance(operation, RollingOperation)
                       for operation in operations])
    if not has_rolling:
        return filters

    dim0 = dimensions[0]
    filters_on_dim0 = [filter_
                       for filter_ in filters
                       if isinstance(filter_, RangeFilter)
                       and str(filter_.definition.term) == str(dim0.definition)]
    if not 0 < len(filters_on_dim0):
        return filters

    max_rolling_period = max(operation.window
                             for operation in operations
                             if isinstance(operation, RollingOperation))

    for filter_ in filters_on_dim0:
        # Monkey patch the update start date on the date filter
        args = {dim0.interval + 's': max_rolling_period} \
            if 'quarter' != dim0.interval \
            else {'months': max_rolling_period * 3}
        filter_.definition.start.value -= relativedelta(**args)

    return filters


def adjust_dataframe_for_rolling_window(operations, data_frame):
    """
    This function adjusts the resulting data frame after executing a slicer query with a rolling operation. If there is
    a date dimension in the first level of the data frame's index and a rolling operation is applied, it will slice the
    dates following the max window to remove it. This way, the adjustment of date filters applied in
    #adjust_daterange_filter_for_rolling_window are removed from the data frame but also in case there are no filters,
    the first few date data points will be removed where the rolling window cannot be calculated.

    :param operations:
    :param data_frame:
    :return:
    """
    has_rolling = any([isinstance(operation, RollingOperation)
                       for operation in operations])
    if not has_rolling:
        return data_frame

    max_rolling_period = max(operation.window
                             for operation in operations
                             if isinstance(operation, RollingOperation))

    if isinstance(data_frame.index, pd.DatetimeIndex):
        return data_frame.iloc[max_rolling_period - 1:]

    if isinstance(data_frame.index, pd.MultiIndex) \
          and isinstance(data_frame.index.levels[0], pd.DatetimeIndex):
        num_levels = len(data_frame.index.levels)

        return data_frame.groupby(level=list(range(1, num_levels))) \
            .apply(lambda df: df.iloc[max_rolling_period - 1:]) \
            .reset_index(level=list(range(num_levels - 1)), drop=True)

    return data_frame


def apply_to_query_args(database, table, joins, dimensions, metrics, operations, filters, references, orders):
    filters = adjust_daterange_filter_for_rolling_window(dimensions, operations, filters)
    return (database, table, joins, dimensions, metrics, operations, filters, references, orders)


def apply_special_cases(f):
    @functools.wraps(f)
    def wrapper(*args):
        return f(*apply_to_query_args(*args))

    return wrapper


def apply_operations_to_data_frame(operations, data_frame):
    data_frame = adjust_dataframe_for_rolling_window(operations, data_frame)

    return data_frame



import pandas as pd

from fireant import (
    ContinuousDimension,
    Metric,
    UniqueDimension,
    formats,
    utils,
)
from fireant.formats import NULL_VALUE
from fireant.utils import (
    MAX_NUMBER,
    MAX_STRING,
    MAX_TIMESTAMP,
    format_dimension_key,
    format_metric_key,
)
from .base import (
    TransformableWidget,
)
from .helpers import (
    dimensional_metric_label,
    extract_display_values,
)
from ..references import (
    reference_key,
    reference_label,
)


def _render_dimension_cell(dimension_value: str, display_values: dict):
    """
    Renders a table cell in a dimension column.

    :param dimension_value:
        The raw value for the table cell.

    :param display_values:
        The display value mapped from the raw value.

    :return:
        A dict with the keys value and possible display.
    """
    dimension_cell = {'value': formats.dimension_value(dimension_value)}

    if display_values is not None:
        if dimension_value in {MAX_STRING, MAX_NUMBER, MAX_TIMESTAMP}:
            dimension_cell['display'] = 'Totals'
        elif pd.isnull(dimension_value):
            dimension_cell['display'] = NULL_VALUE
        else:
            display_value = display_values.get(dimension_value, dimension_value)
            dimension_cell['display'] = formats.dimension_value(display_value)

    return dimension_cell


def _render_dimensional_metric_cell(row_data: pd.Series, metric: Metric):
    """
    Renders a table cell in a metric column for pivoted tables where there are two or more dimensions. This function
    is recursive to traverse multi-dimensional indices.

    :param row_data:
        A series containing the value for the metric and it's index (for the dimension values).

    :param metric:
        A reference to the slicer metric to access the display formatting.

    :return:
        A deep dict in a tree structure with keys matching each dimension level. The top level will have keys matching
        the first level of dimension values, and the next level will contain the next level of dimension values, for as
        many index levels as there are. The last level will contain the return value of `_format_metric_cell`.
    """
    level = {}

    # Group by the last dimension, drop it, and fill the dict with either the raw metric values or the next level of
    # dicts.
    for key, next_row in row_data.groupby(level=1):
        next_row.reset_index(level=1, drop=True, inplace=True)

        df_key = format_metric_key(metric.key)
        level[key] = _render_dimensional_metric_cell(next_row, metric) \
            if isinstance(next_row.index, pd.MultiIndex) \
            else _format_metric_cell(next_row[df_key], metric)

    return level


def _format_metric_cell(value, metric):
    """
    Renders a table cell in a metric column for non-pivoted tables.

    :param value:
        The raw value of the metric.

    :param metric:
        A reference to the slicer metric to access the display formatting.
    :return:
        A dict containing the keys value and display with the raw and display metric values.
    """
    raw_value = formats.metric_value(value)
    return {
        'value': raw_value,
        'display': formats.metric_display(raw_value,
                                          prefix=metric.prefix,
                                          suffix=metric.suffix,
                                          precision=metric.precision)
        if raw_value is not None
        else None
    }


HARD_MAX_COLUMNS = 24


class DataTablesJS(TransformableWidget):
    def __init__(self, metric, *metrics: Metric, pivot=False, max_columns=None):
        super(DataTablesJS, self).__init__(metric, *metrics)
        self.pivot = pivot
        self.max_columns = min(max_columns, HARD_MAX_COLUMNS) \
            if max_columns is not None \
            else HARD_MAX_COLUMNS

    def __repr__(self):
        return '{}({},pivot={})'.format(self.__class__.__name__,
                                        ','.join(str(m) for m in self.items),
                                        self.pivot)

    def transform(self, data_frame, slicer, dimensions, references):
        """
        WRITEME

        :param data_frame:
        :param slicer:
        :param dimensions:
        :return:
        """
        dimension_display_values = extract_display_values(dimensions, data_frame)

        metric_keys = [format_metric_key(reference_key(metric, reference))
                       for metric in self.items
                       for reference in [None] + references]
        data_frame = data_frame[metric_keys]

        pivot_index_to_columns = self.pivot and isinstance(data_frame.index, pd.MultiIndex)
        if pivot_index_to_columns:
            levels = data_frame.index.names[1:]
            data_frame = data_frame \
                .unstack(level=levels) \
                .fillna(value=0)

            dimension_columns = self._dimension_columns(dimensions[:1])

            render_column_label = dimensional_metric_label(dimensions, dimension_display_values)
            metric_columns = self._metric_columns_pivoted(references,
                                                          data_frame.columns,
                                                          render_column_label)

        else:
            dimension_columns = self._dimension_columns(dimensions)
            metric_columns = self._metric_columns(references)

        columns = (dimension_columns + metric_columns)[:self.max_columns]
        data = [self._data_row(dimensions,
                               dimension_values,
                               dimension_display_values,
                               references,
                               row_data)
                for dimension_values, row_data in data_frame.iterrows()]

        return dict(columns=columns, data=data)

    @staticmethod
    def _dimension_columns(dimensions):
        """
        WRITEME

        :param dimensions:
        :return:
        """
        columns = []
        for dimension in dimensions:
            column = dict(title=dimension.label or dimension.key,
                          data=dimension.key,
                          render=dict(_='value'))

            is_cont_dim = isinstance(dimension, ContinuousDimension)
            is_uni_dim_no_display = isinstance(dimension, UniqueDimension) \
                                    and not dimension.has_display_field

            if not is_cont_dim and not is_uni_dim_no_display:
                column['render']['display'] = 'display'

            columns.append(column)

        return columns

    def _metric_columns(self, references):
        """
        WRITEME

        :return:
        """
        columns = []
        for metric in self.items:
            for reference in [None] + references:
                title = reference_label(metric, reference)
                data = reference_key(metric, reference)

                columns.append(dict(title=title,
                                    data=data,
                                    render=dict(_='value', display='display')))
        return columns

    def _metric_columns_pivoted(self, references, df_columns, render_column_label):
        """

        :param references:
        :param df_columns:
        :param render_column_label:
        :return:
        """
        columns = []
        single_metric = 1 == len(self.items)
        for metric in self.items:
            dimension_value_sets = [row[1:]
                                    for row in list(df_columns)]

            for dimension_values in dimension_value_sets:
                for reference in [None] + references:
                    key = reference_key(metric, reference)
                    title = render_column_label(dimension_values, None if single_metric else metric, reference)
                    data = '.'.join([key] + [str(x) for x in dimension_values])

                    columns.append(dict(title=title,
                                        data=data,
                                        render=dict(_='value', display='display')))

        return columns

    def _data_row(self, dimensions, dimension_values, dimension_display_values, references, row_data):
        """
        WRITEME

        :param dimensions:
        :param dimension_values:
        :param dimension_display_values:
        :param row_data:
        :return:
        """
        row = {}

        for dimension, dimension_value in zip(dimensions, utils.wrap_list(dimension_values)):
            df_key = format_dimension_key(dimension.key)
            row[dimension.key] = _render_dimension_cell(dimension_value, dimension_display_values.get(df_key))

        for metric in self.items:
            for reference in [None] + references:
                key = reference_key(metric, reference)
                df_key = format_metric_key(key)

                row[key] = _render_dimensional_metric_cell(row_data, metric) \
                    if isinstance(row_data.index, pd.MultiIndex) \
                    else _format_metric_cell(row_data[df_key], metric)

        return row



class SlicerElement(object):
    """
    The `SlicerElement` class represents an element of the slicer, either a metric or dimension, which contains
    information about such as how to query it from the database.
    """
    def __init__(self, key, label=None, definition=None, display_definition=None):
        """
        :param key:
            The unique identifier of the slicer element, used in the Slicer manager API to reference a defined element.

        :param label:
            A displayable representation of the column.  Defaults to the key capitalized.

        :param definition:
            The definition of the element as a PyPika expression which defines how to query it from the database.

        :param display_definition:
            The definition of the display field for the element as a PyPika expression. Similar to the definition except
            used for querying labels.
        """
        self.key = key
        self.label = label if label is not None else key
        self.definition = definition

        self.display_definition = display_definition

        self.display_key = '{}_display'.format(key) \
            if display_definition is not None \
            else None

    def __repr__(self):
        return self.key

    @property
    def has_display_field(self):
        return getattr(self, 'display_definition', None) is not None

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.key == other.key \
               and str(self.definition) == str(other.definition)

    def __hash__(self):
        return hash('{}({})'.format(self.__class__.__name__, self.key))



from typing import Iterable

from fireant.utils import immutable
from pypika.terms import NullValue
from .base import SlicerElement
from .exceptions import QueryException
from .filters import (
    AntiPatternFilter,
    BooleanFilter,
    ContainsFilter,
    ExcludesFilter,
    PatternFilter,
    RangeFilter,
)
from .intervals import (
    NumericInterval,
    daily,
)


class Dimension(SlicerElement):
    """
    The `Dimension` class represents a dimension in the `Slicer` object.

    :param alias:
        A unique identifier used to identify the metric when writing slicer queries. This value must be unique over
        the metrics in the slicer.

    :param definition:
        A pypika expression which is used to select the value when building SQL queries.

    :param display_definition:
        A pypika expression which is used to select the display value for this dimension.

    :param hyperlink_template:
        A hyperlink template for constructing a URL that can link a value for a dimension to a web page. This is used
        by some transformers such as the ReactTable transformer for displaying hyperlinks.
    """

    def __init__(self, key, label=None, definition=None, display_definition=None, hyperlink_template=None):
        super(Dimension, self).__init__(key, label, definition, display_definition)
        self.is_rollup = False
        self.hyperlink_template = hyperlink_template

    @immutable
    def rollup(self):
        """
        Configures this dimension and all subsequent dimensions in a slicer query to be rolled up to provide the totals.
        This will include an extra value for each pair of dimensions labeled `Totals`. which will include the totals for
        the group.
        """
        self.is_rollup = True

    def __repr__(self):
        return "slicer.dimensions.{}".format(self.key)


class BooleanDimension(Dimension):
    """
    This is a dimension that represents a boolean true/false value.  The expression should always result in a boolean
    value.
    """

    def __init__(self, key, label=None, definition=None, hyperlink_template=None):
        super(BooleanDimension, self).__init__(key,
                                               label=label,
                                               definition=definition,
                                               hyperlink_template=hyperlink_template)

    def is_(self, value: bool):
        """
        Creates a filter to filter a slicer query.

        :param value:
            True or False
        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is True or False.
        """
        return BooleanFilter(self.definition, value)


class PatternFilterableMixin:
    definition = None
    pattern_definition_attribute = 'definition'

    def like(self, pattern, *patterns):
        """
        Creates a filter to filter a slicer query.

        :param pattern:
            A pattern to match against the dimension's display definition.  This pattern is used in the SQL query as
            the `LIKE` expression.
        :param patterns:
            Additional patterns. This is the same as the pattern argument. The function signature is intended to
            syntactically require at least one pattern.
        :return:
            A slicer query filter used to filter a slicer query to results where this dimension's display definition
            matches the pattern.
        """
        definition = getattr(self, self.pattern_definition_attribute)
        return PatternFilter(definition, pattern, *patterns)

    def not_like(self, pattern, *patterns):
        """
        Creates a filter to filter a slicer query.

        :param pattern:
            A pattern to match against the dimension's display definition.  This pattern is used in the SQL query as
            the `NOT LIKE` expression.
        :param patterns:
            Additional patterns. This is the same as the pattern argument. The function signature is intended to
            syntactically require at least one pattern.
        :return:
            A slicer query filter used to filter a slicer query to results where this dimension's display definition
            matches the pattern.
        """
        definition = getattr(self, self.pattern_definition_attribute)
        return AntiPatternFilter(definition, pattern, *patterns)


class CategoricalDimension(PatternFilterableMixin, Dimension):
    """
    This is a dimension that represents an enum-like database field, with a finite list of options to chose from. It
    provides support for configuring a display value for each of the possible values.
    """

    def __init__(self, key, label=None, definition=None, hyperlink_template=None, display_values=()):
        super(CategoricalDimension, self).__init__(key,
                                                   label=label,
                                                   definition=definition,
                                                   hyperlink_template=hyperlink_template)
        self.display_values = dict(display_values)

    def isin(self, values: Iterable):
        """
        Creates a filter to filter a slicer query.

        :param values:
            An iterable of value to constrain the slicer query results by.

        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is one of a set of
            values. Opposite of #notin.
        """
        return ContainsFilter(self.definition, values)

    def notin(self, values):
        """
        Creates a filter to filter a slicer query.

        :param values:
            An iterable of value to constrain the slicer query results by.

        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is *not* one of a set of
            values. Opposite of #isin.
        """
        return ExcludesFilter(self.definition, values)


class _UniqueDimensionBase(PatternFilterableMixin, Dimension):
    def isin(self, values):
        """
        Creates a filter to filter a slicer query.

        :param values:
            An iterable of value to constrain the slicer query results by.

        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is one of a set of
            values. Opposite of #notin.
        """
        return ContainsFilter(self.definition, values)

    def notin(self, values):
        """
        Creates a filter to filter a slicer query.

        :param values:
            An iterable of value to constrain the slicer query results by.

        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is *not* one of a set of
            values. Opposite of #isin.
        """
        return ExcludesFilter(self.definition, values)


class UniqueDimension(_UniqueDimensionBase):
    """
    This is a dimension that represents a field in a database which is a unique identifier, such as a primary/foreign
    key. It provides support for a display value field which is selected and used in the results.
    """

    def __init__(self, key, label=None, definition=None, display_definition=None, hyperlink_template=None):
        super(UniqueDimension, self).__init__(key,
                                              label=label,
                                              definition=definition,
                                              display_definition=display_definition,
                                              hyperlink_template=hyperlink_template)
        if display_definition is not None:
            self.display = DisplayDimension(self)

    @property
    def has_display_field(self):
        return hasattr(self, 'display')

    def like(self, pattern, *patterns):
        if not self.has_display_field:
            raise QueryException('No value set for display_definition.')
        return self.display.like(pattern, *patterns)

    def not_like(self, pattern, *patterns):
        if not self.has_display_field:
            raise QueryException('No value set for display_definition.')
        return self.display.not_like(pattern, *patterns)


class DisplayDimension(_UniqueDimensionBase):
    """
    WRITEME
    """

    def __init__(self, dimension):
        super(DisplayDimension, self).__init__('{}_display'.format(dimension.key),
                                               label=dimension.label,
                                               definition=dimension.display_definition,
                                               hyperlink_template=dimension.hyperlink_template)


class ContinuousDimension(Dimension):
    """
    This is a dimension that represents a field in the database which is a continuous value, such as a decimal, integer,
    or date/time. It requires the use of an interval which is the window over the values.
    """

    def __init__(self, key, label=None, definition=None, hyperlink_template=None,
                 default_interval=NumericInterval(1, 0)):
        super(ContinuousDimension, self).__init__(key,
                                                  label=label,
                                                  definition=definition,
                                                  hyperlink_template=hyperlink_template)
        self.interval = default_interval


class DatetimeDimension(ContinuousDimension):
    """
    A subclass of ContinuousDimension which reflects a date/time data type. Intervals are replaced with time intervals
    such as daily, weekly, annually, etc.  A reference can be used to show a comparison over time such as
    week-over-week or month-over-month.
    """

    def __init__(self, key, label=None, definition=None, hyperlink_template=None, default_interval=daily):
        super(DatetimeDimension, self).__init__(key,
                                                label=label,
                                                definition=definition,
                                                hyperlink_template=hyperlink_template,
                                                default_interval=default_interval)

    @immutable
    def __call__(self, interval):
        """
        When calling a datetime dimension an interval can be supplied:

        .. code-block:: python

            from fireant import weekly

            my_slicer.dimensions.date # Daily interval used as default
            my_slicer.dimensions.date(weekly) # Daily interval used as default

        :param interval:
            An interval to use with the dimension.  See `fireant.intervals`.
        :return:
            A copy of the dimension with the interval set.
        """
        self.interval = interval

    def between(self, start, stop):
        """
        Creates a filter to filter a slicer query.

        :param start:
            The start time of the filter. This is the beginning of the window for which results should be included.
        :param stop:
            The stop time of the filter. This is the end of the window for which results should be included.
        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is between the values
            start and stop.
        """
        return RangeFilter(self.definition, start, stop)


class TotalsDimension(Dimension):
    def __init__(self, dimension):
        totals_definition = NullValue()
        display_definition = totals_definition \
            if dimension.has_display_field \
            else None

        super(TotalsDimension, self).__init__(dimension.key,
                                              label=dimension.label,
                                              definition=totals_definition,
                                              display_definition=display_definition,
                                              hyperlink_template=dimension.hyperlink_template)



from pypika import functions as fn
from pypika.queries import QueryBuilder
from fireant import utils


class Reference(object):
    def __init__(self, dimension, reference_type, delta=False, delta_percent=False):
        self.dimension = dimension

        self.reference_type = reference_type
        self.key = reference_type.key + '_delta_percent' \
            if delta_percent \
            else reference_type.key + '_delta' \
            if delta \
            else reference_type.key

        self.label = reference_type.label + ' Δ%' \
            if delta_percent \
            else reference_type.label + ' Δ' \
            if delta \
            else reference_type.label

        self.time_unit = reference_type.time_unit
        self.interval = reference_type.interval

        self.delta = delta_percent or delta
        self.delta_percent = delta_percent

    def __eq__(self, other):
        return isinstance(self, Reference) \
               and self.dimension == other.dimension \
               and self.key == other.key

    def __hash__(self):
        return hash('reference{}{}'.format(self.key, self.dimension))

    def __repr__(self):
        return '{}({})'.format(self.key, self.dimension.key)


class ReferenceType(object):
    def __init__(self, key, label, time_unit: str, interval: int):
        self.key = key
        self.label = label

        self.time_unit = time_unit
        self.interval = interval

    def __call__(self, dimension, delta=False, delta_percent=False):
        return Reference(dimension, self, delta=delta, delta_percent=delta_percent)

    def __eq__(self, other):
        return isinstance(self, ReferenceType) \
               and self.time_unit == other.time_unit \
               and self.interval == other.interval

    def __hash__(self):
        return hash('reference' + self.key)


DayOverDay = ReferenceType('dod', 'DoD', 'day', 1)
WeekOverWeek = ReferenceType('wow', 'WoW', 'week', 1)
MonthOverMonth = ReferenceType('mom', 'MoM', 'month', 1)
QuarterOverQuarter = ReferenceType('qoq', 'QoQ', 'quarter', 1)
YearOverYear = ReferenceType('yoy', 'YoY', 'year', 1)


def reference_key(metric, reference):
    """
    Format a metric key for a reference.

    :return:
        A string that is used as the key for a reference metric.
    """
    key = metric.key

    if reference is not None:
        return '{}_{}'.format(key, reference.key)

    return key


def reference_label(metric, reference):
    """
    Format a metric label for a reference.

    :return:
        A string that is used as the display value for a reference metric.
    """
    label = metric.label or metric.key

    if reference is not None:
        return '{} ({})'.format(label, reference.label)

    return label


def reference_term(reference: Reference,
                   original_query: QueryBuilder,
                   ref_query: QueryBuilder):
    """
    Part of query building. Given a reference, the original slicer query, and the ref query, creates the pypika for
    the reference that should be selected in the reference container query.

    :param reference:
    :param original_query:
    :param ref_query:
    :return:
    """

    def original_field(metric):
        return original_query.field(utils.format_metric_key(metric.key))

    def ref_field(metric):
        return ref_query.field(utils.format_metric_key(metric.key))

    if reference.delta:
        if reference.delta_percent:
            return lambda metric: (original_field(metric) - ref_field(metric)) \
                                  * \
                                  (100 / fn.NullIf(ref_field(metric), 0))

        return lambda metric: original_field(metric) - ref_field(metric)

    return ref_field



import codecs
import os
import re

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='fireant',
    version=find_version("fireant", "__init__.py"),

    author='KAYAK, GmbH',
    author_email='bandit@kayak.com',

    # License
    license='Apache License Version 2.0',

    packages=['fireant',
              'fireant.database',
              'fireant.slicer',
              'fireant.slicer.queries',
              'fireant.slicer.widgets'],

    # Include additional files into the package
    include_package_data=True,

    # Details
    url='https://github.com/kayak/fireant',

    description='Data analysis tool for Python and Jupyter Notebooks',
    long_description="Fireant is a tool which aims to simplify the process of accessing analytical data and building "
                     "dashboards and reports.  It eliminates the need for handwritten queries or code to process data. "
                     "The main feature, the slicer, is a schema that is used to configure metrics, the data being "
                     "visualized, and dimensions, the spatiality over which metrics can be extended and grouped. "
                     "Included in the Slicer is a Slicer Manager API which exposes methods to render tables and charts "
                     "in a variety of different formats.\n\n"

                     "Fireant is a great compliment to Jupter notebooks using matplotlib and pandas. It can be used in "
                     "web projects with Highcharts and Datatables.js.  It can even just provide a layer of abstraction "
                     "for accessing analytical data akin to ORM libraries in software projects.",

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: PL/SQL',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords=('fireant python query builder querybuilder sql mysql postgres psql oracle vertica aggregated '
              'relational database rdbms business analytics bi data science analysis pandas'),

    install_requires=required,
    tests_require=[
        'mock'
    ],
    extras_require={
        'vertica': ['vertica-python==0.7.3'],
        'mysql': ['pymysql==0.8.0'],
        'redshift': ['psycopg2==2.7.3.2'],
        'postgresql': ['psycopg2==2.7.3.2'],
        'matplotlib': ['matplotlib'],
    },

    test_suite='fireant.tests',
)



import numpy as np
import pandas as pd
from datetime import (
    date,
    datetime,
    time,
)

from fireant.utils import (
    MAX_NUMBER,
    MAX_STRING,
    MAX_TIMESTAMP,
)

INF_VALUE = "Inf"
NULL_VALUE = 'null'
NAN_VALUE = 'NaN'
TOTALS_VALUE = 'totals'
RAW_VALUE = 'raw'

NO_TIME = time(0)

epoch = np.datetime64(datetime.utcfromtimestamp(0))
milliseconds = np.timedelta64(1, 'ms')


def date_as_millis(value):
    if not isinstance(value, date):
        value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
    return int(1000 * value.timestamp())


def coerce_type(value):
    if isinstance(value, date):
        return value

    for type_cast in (int, float):
        try:
            return type_cast(value)
        except:
            pass

    # Should never be any real NaNs or INFs at this point, so if that's the value, it's meant to be that.
    if value.lower() in ['nan', 'inf']:
        return value

    if NULL_VALUE == value:
        return None
    if 'True' == value:
        return True
    if 'False' == value:
        return False

    return value


def dimension_value(value):
    """
    Format a dimension value. This will coerce the raw string or date values into a proper primitive value like a
    string, float, or int.

    :param value:
        The raw str or datetime value
    :param str_date:
        When True, dates and datetimes will be converted to ISO strings. The time is omitted for dates. When False, the
        datetime will be converted to a POSIX timestamp (millis-since-epoch).
    """
    if value in {MAX_STRING, MAX_NUMBER, MAX_TIMESTAMP}:
        return 'Totals'

    if pd.isnull(value):
        return NULL_VALUE

    if isinstance(value, date):
        if not hasattr(value, 'time') or value.time() == NO_TIME:
            return value.strftime('%Y-%m-%d')
        else:
            return value.strftime('%Y-%m-%d %H:%M:%S')

    return coerce_type(value)


def metric_value(value):
    """
    Converts a raw metric value into a safe type. This will change dates into strings, NaNs into Nones, and np types
    into their corresponding python types.

    :param value:
        The raw metric value.
    """
    if isinstance(value, date):
        if not hasattr(value, 'time') or value.time() == NO_TIME:
            return value.strftime('%Y-%m-%d')
        else:
            return value.strftime('%Y-%m-%dT%H:%M:%S')

    if value is None or pd.isnull(value):
        return None

    if isinstance(value, float):
        if np.isinf(value):
            return None
        return float(value)

    if isinstance(value, np.int64):
        # Cannot transform np.int64 to json
        return int(value)

    return value


def metric_display(value, prefix=None, suffix=None, precision=None):
    """
    Converts a metric value into the display value by applying formatting.

    :param value:
        The raw metric value.
    :param prefix:
        An optional prefix.
    :param suffix:
        An optional suffix.
    :param precision:
        The decimal precision, the number of decimal places to round to.
    :return:
        A formatted string containing the display value for the metric.
    """
    if pd.isnull(value):
        value = NULL_VALUE

    if value in {np.inf, -np.inf, INF_VALUE}:
        return INF_VALUE

    if value in (NULL_VALUE, NAN_VALUE):
        return NULL_VALUE

    if isinstance(value, bool):
        value = str(value).lower()

    if prefix == '$' and isinstance(value, (float, int)) and value < 0:
        value = -value
        prefix = '-$'

    if isinstance(value, float):
        if precision is not None:
            value = '{:,.{precision}f}'.format(value, precision=precision)

        elif value.is_integer():
            value = '{:,.0f}'.format(value)

        else:
            # Stripping trailing zeros is necessary because %f can add them if no precision is set
            value = '{:,f}'.format(value).rstrip('.0')

    if isinstance(value, int):
        value = '{:,.0f}'.format(value)

    return '{prefix}{value}{suffix}'.format(
          prefix=prefix or '',
          value=str(value),
          suffix=suffix or '',
    )



import numpy as np
import pandas as pd

from fireant.utils import format_metric_key
from fireant.slicer.references import reference_key
from .metrics import Metric


def _extract_key_or_arg(data_frame, key):
    return data_frame[key] \
        if key in data_frame \
        else key


class Operation(object):
    """
    The `Operation` class represents an operation in the `Slicer` API.
    """

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        raise NotImplementedError()

    @property
    def operations(self):
        return []


class _BaseOperation(Operation):
    def __init__(self, key, label, prefix=None, suffix=None, precision=None):
        self.key = key
        self.label = label
        self.prefix = prefix
        self.suffix = suffix
        self.precision = precision

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        raise NotImplementedError()

    @property
    def operations(self):
        raise NotImplementedError()

    def _group_levels(self, index):
        """
        Get the index levels that need to be grouped. This is to avoid apply the cumulative function across separate
        dimensions. Only the first dimension should be accumulated across.

        :param index:
        :return:
        """
        return index.names[1:]


class _Cumulative(_BaseOperation):
    def __init__(self, arg):
        super(_Cumulative, self).__init__(
              key='{}({})'.format(self.__class__.__name__.lower(),
                                  getattr(arg, 'key', arg)),
              label='{}({})'.format(self.__class__.__name__,
                                    getattr(arg, 'label', arg)),
              prefix=getattr(arg, 'prefix'),
              suffix=getattr(arg, 'suffix'),
              precision=getattr(arg, 'precision'),
        )

        self.arg = arg

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        return [metric
                for metric in [self.arg]
                if isinstance(metric, Metric)]

    @property
    def operations(self):
        return [op_and_children
                for operation in [self.arg]
                if isinstance(operation, Operation)
                for op_and_children in [operation] + operation.operations]

    def __repr__(self):
        return self.key


class CumSum(_Cumulative):
    def apply(self, data_frame, reference):
        df_key = format_metric_key(reference_key(self.arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .cumsum()

        return data_frame[df_key].cumsum()


class CumProd(_Cumulative):
    def apply(self, data_frame, reference):
        df_key = format_metric_key(reference_key(self.arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .cumprod()

        return data_frame[df_key].cumprod()


class CumMean(_Cumulative):
    @staticmethod
    def cummean(x):
        return x.cumsum() / np.arange(1, len(x) + 1)

    def apply(self, data_frame, reference):
        df_key = format_metric_key(reference_key(self.arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .apply(self.cummean)

        return self.cummean(data_frame[df_key])


class RollingOperation(_BaseOperation):
    def __init__(self, arg, window, min_periods=None):
        super(RollingOperation, self).__init__(
              key='{}({})'.format(self.__class__.__name__.lower(),
                                  getattr(arg, 'key', arg)),
              label='{}({})'.format(self.__class__.__name__,
                                    getattr(arg, 'label', arg)),
              prefix=getattr(arg, 'prefix'),
              suffix=getattr(arg, 'suffix'),
              precision=getattr(arg, 'precision'),
        )

        self.arg = arg
        self.window = window
        self.min_periods = min_periods

    def _should_adjust(self, other_operations):
        # Need to figure out if this rolling operation is has the largest window, and if it's the first of multiple
        # rolling operations if there are more than one operation sharing the largest window.
        first_max_rolling = list(sorted(other_operations, key=lambda operation: operation.window))[0]

        return first_max_rolling is self

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        return [metric
                for metric in [self.arg]
                if isinstance(metric, Metric)]

    @property
    def operations(self):
        return [op_and_children
                for operation in [self.arg]
                if isinstance(operation, Operation)
                for op_and_children in [operation] + operation.operations]


class RollingMean(RollingOperation):
    def rolling_mean(self, x):
        return x.rolling(self.window, self.min_periods).mean()

    def apply(self, data_frame, reference):
        df_key = format_metric_key(reference_key(self.arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .apply(self.rolling_mean)

        return self.rolling_mean(data_frame[df_key])



import numpy as np

from fireant import utils
from fireant.formats import (
    INF_VALUE,
    NULL_VALUE,
)
from fireant.utils import (
    MAX_NUMBER,
    MAX_STRING,
    MAX_TIMESTAMP,
)
from ..references import reference_label

TOTALS_MARKERS = {
    MAX_NUMBER,
    MAX_STRING,
    MAX_TIMESTAMP,
}


def extract_display_values(dimensions, data_frame):
    """
    Retrieves the display values for each dimension.

    For UniqueDimension, this will retrieve the display values from the data frame containing the data from the slicer
    query. For CategoricalDimension, the values are retrieved from the set of display values configured in the slicer.

    :param dimensions:
        A list of dimensions present in a slicer query.
    :param data_frame:
        The data frame containing the data result of the slicer query.
    :return:
        A dict containing keys for dimensions with display values (If there are no display values then the
        dimension's key will not be present). The value of the dict will be either a dict or a data frame where the
        display value can be accessed using the display value as the key.
    """
    display_values = {}

    for dimension in dimensions:
        key = utils.format_dimension_key(dimension.key)

        if hasattr(dimension, 'display_values'):
            display_values[key] = dimension.display_values

        elif dimension.has_display_field:
            f_display_key = utils.format_dimension_key(dimension.display_key)

            display_values[key] = data_frame[f_display_key] \
                .groupby(level=key) \
                .first() \
                .fillna(value=NULL_VALUE) \
                .replace([np.inf, -np.inf], INF_VALUE)

    return display_values


def dimensional_metric_label(dimensions, dimension_display_values):
    """
    Creates a callback function for rendering series labels.

    :param dimensions:
        A list of fireant.Dimension which is being rendered.

    :param dimension_display_values:
        A dictionary containing key-value pairs for each dimension.
    :return:
        a callback function which renders a label for a metric, reference, and list of dimension values.
    """

    def render_series_label(dimension_values, metric=None, reference=None):
        """
        Returns a string label for a metric, reference, and set of values for zero or more dimensions.

        :param metric:
            an instance of fireant.Metric
        :param reference:
            an instance of fireant.Reference
        :param dimension_values:
            a tuple of dimension values. Can be zero-length or longer.
        :return:
        """
        num_used_dimensions = len(dimensions) - len(dimension_values)
        used_dimensions = dimensions[num_used_dimensions:]

        dimension_values = utils.wrap_list(dimension_values)
        dimension_labels = [utils.getdeepattr(dimension_display_values,
                                              (utils.format_dimension_key(dimension.key), dimension_value),
                                              dimension_value)
                            if dimension_value not in TOTALS_MARKERS
                            else 'Totals'
                            for dimension, dimension_value in zip(used_dimensions, dimension_values)]

        label = ", ".join([str(label) for label in dimension_labels])

        if metric is None:
            if reference is not None:
                return '{} ({})'.format(label, reference.label)
            return label

        if dimension_labels:
            return '{} ({})'.format(reference_label(metric, reference),
                                    label)

        return reference_label(metric, reference)

    return render_series_label



from .dimensions import (
    BooleanDimension,
    CategoricalDimension,
    ContinuousDimension,
    DatetimeDimension,
    Dimension,
    DisplayDimension,
    UniqueDimension,
)
from .exceptions import (
    QueryException,
    SlicerException,
)
from .intervals import (
    DatetimeInterval,
    NumericInterval,
    annually,
    daily,
    hourly,
    monthly,
    quarterly,
    weekly,
)
from .joins import Join
from .metrics import Metric
from .operations import (
    CumMean,
    CumProd,
    CumSum,
    Operation,
    RollingMean,
)
from .references import (
    DayOverDay,
    MonthOverMonth,
    QuarterOverQuarter,
    Reference,
    WeekOverWeek,
    YearOverYear,
)
from .slicers import Slicer



import re
from collections import OrderedDict

import numpy as np
import pandas as pd

from fireant.formats import (
    INF_VALUE,
    NULL_VALUE,
    RAW_VALUE,
    TOTALS_VALUE,
    metric_display,
    metric_value,
)
from fireant.utils import (
    format_metric_key,
    getdeepattr,
    setdeepattr,
)
from .pandas import Pandas
from ..dimensions import (
    DatetimeDimension,
    Dimension,
)
from ..intervals import (
    annually,
    daily,
    hourly,
    monthly,
    quarterly,
    weekly,
)
from ..metrics import Metric
from ..references import (
    reference_key,
    reference_label,
)

DATE_FORMATS = {
    hourly: '%Y-%m-%d %h',
    daily: '%Y-%m-%d',
    weekly: '%Y-%m',
    monthly: '%Y',
    quarterly: '%Y-%q',
    annually: '%Y',
}
from fireant.utils import (
    MAX_NUMBER,
    MAX_STRING,
    MAX_TIMESTAMP,
    format_dimension_key,
    wrap_list,
)

TOTALS_LABEL = 'Totals'
metrics = Dimension('metrics', '')
metrics_dimension_key = format_dimension_key(metrics.key)


def map_index_level(index, level, func):
    # If the index is empty, do not do anything
    if 0 == index.size:
        return index

    if isinstance(index, pd.MultiIndex):
        values = index.levels[level]
        return index.set_levels(values.map(func), level)

    assert level == 0

    return index.map(func)


class ReferenceItem:
    def __init__(self, item, reference):
        if reference is None:
            self.key = item.key
            self.label = item.label
        else:
            self.key = reference_key(item, reference)
            self.label = reference_label(item, reference)

        self.prefix = item.prefix
        self.suffix = item.suffix
        self.precision = item.precision


class TotalsItem:
    key = TOTALS_VALUE
    label = TOTALS_LABEL
    prefix = suffix = precision = None


class ReactTable(Pandas):
    """
    This component does not work with react-table out of the box, some customization is needed in order to work with
    the transformed data.

    .. code-block:: jsx

        // A Custom TdComponent implementation is required by Fireant in order to render display values
        const TdComponent = ({
                               toggleSort,
                               className,
                               children,
                               ...rest
                             }) =>
            <div className={classNames('rt-td', className)} role="gridcell" {...rest}>
                {_.get(children, 'display', children.raw) || <span>&nbsp;</span>}
            </div>;

        const FireantReactTable = ({
                                config, // The payload from fireant
                              }) =>
            <ReactTable columns={config.columns}
                        data={config.data}
                        minRows={0}

                        TdComponent={ DashmoreTdComponent}
                        defaultSortMethod={(a, b, desc) => ReactTableDefaults.defaultSortMethod(a.raw, b.raw, desc)}>
            </ReactTable>;
    """

    def __init__(self, metric, *metrics: Metric, pivot=(), transpose=False, sort=None, ascending=None,
                 max_columns=None):
        super(ReactTable, self).__init__(metric, *metrics,
                                         pivot=pivot,
                                         transpose=transpose,
                                         sort=sort,
                                         ascending=ascending,
                                         max_columns=max_columns)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ','.join(str(m) for m in self.items))

    @staticmethod
    def map_display_values(df, dimensions):
        """
        Creates a mapping for dimension values to their display values.

        :param df:
            The result data set that is being transformed.
        :param dimensions:
            The list of dimensions included in the query that created the result data set df.
        :return:
            A tree-structure dict with two levels of depth. The top level dict has keys for each dimension's display
            key. The lower level dict has keys for each raw dimension value and values which are the display value.
        """
        dimension_display_values = {}

        for dimension in dimensions:
            f_dimension_key = format_dimension_key(dimension.key)

            if dimension.has_display_field:
                f_display_key = format_dimension_key(dimension.display_key)

                dimension_display_values[f_dimension_key] = \
                    df[f_display_key].groupby(f_dimension_key).first().to_dict()

                del df[f_display_key]

            if hasattr(dimension, 'display_values'):
                dimension_display_values[f_dimension_key] = dimension.display_values

        return dimension_display_values

    @staticmethod
    def map_hyperlink_templates(df, dimensions):
        """
        Creates a mapping for each dimension to it's hyperlink template if it is possible to create the hyperlink
        template for it.

        The hyperlink template is a URL-like string containing curley braces enclosing dimension keys: `{dimension}`.
        While rendering this widget, the dimension key placeholders need to be replaced with the dimension values for
        that row.

        :param df:
            The result data set that is being transformed. The data frame SHOULD be pivoted/transposed if that step is
            required, before calling this function, in order to prevent the template from being included for the
            dimension if one of the required dimensions is pivoted.
        :param dimensions:
            The list of dimensions included in the query that created the result data set df.
        :return:
            A dict with the dimension key as the key and the hyperlink template as the value. Templates will only be
            included if it will be possible to fill in the required parameters.
        """
        hyperlink_templates = {}
        pattern = re.compile(r'{[^{}]+}')

        for dimension in dimensions:
            hyperlink_template = dimension.hyperlink_template
            if hyperlink_template is None:
                continue

            required_hyperlink_parameters = [format_dimension_key(argument[1:-1])
                                             for argument in pattern.findall(hyperlink_template)]

            # Check that all of the required dimensions are in the result data set. Only include the hyperlink template
            # in the return value of this function if all are present.
            unavailable_hyperlink_parameters = set(required_hyperlink_parameters) & set(df.index.names)
            if not unavailable_hyperlink_parameters:
                continue

            # replace the dimension keys with the formatted values. This will come in handy later when replacing the
            # actual values
            hyperlink_template = hyperlink_template.format(**{
                argument[3:]: '{' + argument + '}'
                for argument in required_hyperlink_parameters
            })

            f_dimension_key = format_dimension_key(dimension.key)
            hyperlink_templates[f_dimension_key] = hyperlink_template

        return hyperlink_templates

    @staticmethod
    def format_data_frame(data_frame, dimensions):
        """
        This function prepares the raw data frame for transformation by formatting dates in the index and removing any
        remaining NaN/NaT values. It also names the column as metrics so that it can be treated like a dimension level.

        :param data_frame:
            The result set data frame
        :param dimensions:
        :return:
        """
        for i, dimension in enumerate(dimensions):
            if isinstance(dimension, DatetimeDimension):
                date_format = DATE_FORMATS.get(dimension.interval, DATE_FORMATS[daily])

                def format_datetime(dt):
                    if MAX_TIMESTAMP == dt:
                        return TOTALS_LABEL
                    return dt.strftime(date_format)

                data_frame.index = map_index_level(data_frame.index, i, format_datetime)

        data_frame.columns.name = metrics_dimension_key

    @staticmethod
    def transform_dimension_column_headers(data_frame, dimensions):
        """
        Convert the un-pivoted dimensions into ReactTable column header definitions.

        :param data_frame:
            The result set data frame
        :param dimensions:
            A list of dimensions in the data frame that are part of the index
        :return:
            A list of column header definitions with the following structure.


        .. code-block:: jsx

            columns = [{
              Header: 'Column A',
              accessor: 'a',
            }, {
              Header: 'Column B',
              accessor: 'b',
            }]
        """
        dimension_map = {format_dimension_key(d.key): d
                         for d in dimensions + [metrics]}

        columns = []
        if not isinstance(data_frame.index, pd.MultiIndex) and data_frame.index.name is None:
            return columns

        for f_dimension_key in data_frame.index.names:
            dimension = dimension_map[f_dimension_key]

            columns.append({
                'Header': getattr(dimension, 'label', dimension.key),
                'accessor': f_dimension_key,
            })

        return columns

    @staticmethod
    def transform_metric_column_headers(data_frame, item_map, dimension_display_values):
        """
        Convert the metrics into ReactTable column header definitions. This includes any pivoted dimensions, which will
        result in multiple rows of headers.

        :param data_frame:
            The result set data frame
        :param item_map:
            A map to find metrics/operations based on their keys found in the data frame.
        :param dimension_display_values:
            A map for finding display values for dimensions based on their key and value.
        :return:
            A list of column header definitions with the following structure.

        .. code-block:: jsx

            columns = [{
              Header: 'Column A',
              columns: [{
                Header: 'SubColumn A.0',
                accessor: 'a.0',
              }, {
                Header: 'SubColumn A.1',
                accessor: 'a.1',
              }]
            }, {
              Header: 'Column B',
              columns: [
                ...
              ]
            }]
        """

        def get_header(column_value, f_dimension_key, is_totals):
            if f_dimension_key == metrics_dimension_key or is_totals:
                item = item_map[column_value]
                return getattr(item, 'label', item.key)

            return getdeepattr(dimension_display_values, (f_dimension_key, column_value), column_value)

        def _make_columns(columns_frame, previous_levels=()):
            """
            This function recursively creates the individual column definitions for React Table with the above tree
            structure depending on how many index levels there are in the columns.

            :param columns_frame:
                A data frame representing the columns of the result set data frame.
            :param previous_levels:
                A tuple containing the higher level index level values used for building the data accessor path
            """
            f_dimension_key = columns_frame.index.names[0]

            # Group the columns if they are multi-index so we can get the proper sub-column values. This will yield
            # one group per dimension value with the group data frame containing only the relevant sub-columns
            groups = columns_frame.groupby(level=0) \
                if isinstance(columns_frame.index, pd.MultiIndex) else \
                [(level, None) for level in columns_frame.index]

            columns = []
            for column_value, group in groups:
                is_totals = column_value in {MAX_STRING, MAX_NUMBER, TOTALS_LABEL}

                # All column definitions have a header
                column = {'Header': get_header(column_value, f_dimension_key, is_totals)}

                levels = previous_levels + (column_value,)
                if group is not None:
                    # If there is a group, then drop this index level from the group data frame and recurse to build
                    # sub column definitions
                    next_level_df = group.reset_index(level=0, drop=True)
                    column['columns'] = _make_columns(next_level_df, levels)

                else:
                    column['accessor'] = '.'.join(levels)

                if is_totals:
                    column['className'] = 'fireant-totals'

                columns.append(column)

            return columns

        column_frame = data_frame.columns.to_frame()
        return _make_columns(column_frame)

    @classmethod
    def transform_data_row_index(cls, index_values, dimension_display_values, dimension_hyperlink_templates):
        # Add the index to the row
        row = {}
        for key, value in index_values.items():
            if key is None:
                continue

            data = {RAW_VALUE: metric_value(value)}

            # Try to find a display value for the item. If this is a metric the raw value is replaced with the
            # display value because there is no raw value for a metric label
            display = getdeepattr(dimension_display_values, (key, value))
            if display is not None:
                data['display'] = display

            # If the dimension has a hyperlink template, then apply the template by formatting it with the dimension
            # values for this row. The values contained in `index_values` will always contain all of the required values
            # at this point, otherwise the hyperlink template will not be included.
            if key in dimension_hyperlink_templates:
                data['hyperlink'] = dimension_hyperlink_templates[key].format(**index_values)

            row[key] = data

        return row

    @classmethod
    def transform_data_row_values(cls, series, item_map):
        # Add the values to the row
        row = {}
        for key, value in series.iteritems():
            key = wrap_list(key)
            value = metric_value(value)
            data = {RAW_VALUE: metric_value(value)}

            # Try to find a display value for the item
            item = item_map.get(key[0])
            display = metric_display(value,
                                     getattr(item, 'prefix', None),
                                     getattr(item, 'suffix', None),
                                     getattr(item, 'precision', None))

            if display is not None:
                data['display'] = display

            setdeepattr(row, key, data)

        return row

    @classmethod
    def transform_data(cls, data_frame, item_map, dimension_display_values, dimension_hyperlink_templates):
        """
        Builds a list of dicts containing the data for ReactTable. This aligns with the accessors set by
        #transform_dimension_column_headers and #transform_metric_column_headers

        :param data_frame:
            The result set data frame
        :param item_map:
            A map to find metrics/operations based on their keys found in the data frame.
        :param dimension_display_values:
            A map for finding display values for dimensions based on their key and value.
        :param dimension_hyperlink_templates:
        """
        index_names = data_frame.index.names

        rows = []

        for index, series in data_frame.iterrows():
            if not isinstance(index, tuple):
                index = (index,)

            # Get a list of values from the index. These can be metrics or dimensions so it checks in the item map if
            # there is a display value for the value
            index = [item
                     if item not in item_map
                     else getattr(item_map[item], 'label', item_map[item].key)
                     for item in index]
            index_values = OrderedDict(zip(index_names, index))

            index_cols = cls.transform_data_row_index(index_values,
                                                      dimension_display_values,
                                                      dimension_hyperlink_templates)
            value_cols = cls.transform_data_row_values(series, item_map)

            row = {}
            row.update(index_cols)
            row.update(value_cols)
            rows.append(row)

        return rows

    def transform(self, data_frame, slicer, dimensions, references):
        """
        Transforms a data frame into a format for ReactTable. This is an object containing attributes `columns` and
        `data` which align with the props in ReactTable with the same name.

        :param data_frame:
            The result set data frame
        :param slicer:
            The slicer that generated the data query
        :param dimensions:
            A list of dimensions that were selected in the data query
        :param references:
            A list of references that were selected in the data query
        :return:
            An dict containing attributes `columns` and `data` which align with the props in ReactTable with the same
            names.
        """
        df_dimension_columns = [format_dimension_key(d.display_key)
                                for d in dimensions
                                if d.has_display_field]
        item_map = OrderedDict([(format_metric_key(reference_key(i, reference)), ReferenceItem(i, reference))
                                for i in self.items
                                for reference in [None] + references])
        df_metric_columns = list(item_map.keys())

        # Add an extra item to map the totals markers to it's label
        item_map[MAX_NUMBER] = TotalsItem
        item_map[MAX_STRING] = TotalsItem
        item_map[TOTALS_LABEL] = TotalsItem

        df = data_frame[df_dimension_columns + df_metric_columns].copy()

        dimension_display_values = self.map_display_values(df, dimensions)

        self.format_data_frame(df, dimensions)

        dimension_keys = [format_dimension_key(dimension.key) for dimension in self.pivot]
        df = self.pivot_data_frame(df, dimension_keys, self.transpose) \
            .fillna(value=NULL_VALUE) \
            .replace([np.inf, -np.inf], INF_VALUE)

        dimension_hyperlink_templates = self.map_hyperlink_templates(df, dimensions)

        dimension_columns = self.transform_dimension_column_headers(df, dimensions)
        metric_columns = self.transform_metric_column_headers(df, item_map, dimension_display_values)
        data = self.transform_data(df, item_map, dimension_display_values, dimension_hyperlink_templates)

        return {
            'columns': dimension_columns + metric_columns,
            'data': data,
        }
